/*!
        @file    $Id:: field_G_imp.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru) 
                 $LastChangedBy: matufuru $

        @date    $LastChangedDate:: 2014-06-02 23:54:18 #$

        @version $LastChangedRevision: 1054 $
*/

#include "field_G.h"

#include "bridgeIO.h"
using Bridge::vout;

#include "threadManager_OpenMP.h"

#if defined USE_GROUP_SU3
#include "field_G_imp_SU3.inc"
#elif defined USE_GROUP_SU2
#include "field_G_imp_SU2.inc"
#elif defined USE_GROUP_SU_N
#include "field_G_imp_SU_N.inc"
#endif

//====================================================================
void Field_G::check()
{
  assert(NC == CommonParameters::Nc());

  check_Nc();
}

//====================================================================
void Field_G::set_unity()
{
  Mat_SU_N ut(m_Nc);
  ut.unit();

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  for(int mu = 0; mu < m_Nex; ++mu){
   for(int site = is; site < is+ns; ++site){
     set_mat(site, mu, ut);
   }
  }

}

//====================================================================
void Field_G::reunit()
{
  Mat_SU_N ut(m_Nc);

  //  int nth = ThreadManager_OpenMP::get_num_threads();
  //  int ith = ThreadManager_OpenMP::get_thread_id();
  int nth = 1;
  int ith = 0;

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  for(int mu = 0; mu < m_Nex; ++mu){
   for(int site = is; site < is+ns; ++site){
     mat(ut, site, mu);
     ut.reunit();
     set_mat(site, mu, ut);
   }
  }

}

//====================================================================
void Field_G::mult_Field_Gnn(int ex, const Field_G& u1, int ex1,
                                     const Field_G& u2, int ex2)
{
  assert(ex < m_Nex);
  assert(ex1 < u1.nex());
  assert(ex2 < u2.nex());
  assert(u1.nvol() == m_Nvol);
  assert(u2.nvol() == m_Nvol);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));
  double* g1 = const_cast<Field_G*>(&u1)->ptr(m_Nin*(is + m_Nvol*ex1));
  double* g2 = const_cast<Field_G*>(&u2)->ptr(m_Nin*(is + m_Nvol*ex2));

  int Nc2  = 2 * m_Nc;
  int Ndf  = Nc2 * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int ic2 = 0; ic2 < m_Nc; ++ic2){
    for(int ic1 = 0; ic1 < m_Nc; ++ic1){
      int jg2 = ic2 * 2   + ig;
      int jg1 = ic1 * Nc2 + ig;
      g[2*ic2   + Nc2*ic1 + ig] = mult_Gnn_r(&g1[jg1], &g2[jg2], m_Nc);
      g[2*ic2+1 + Nc2*ic1 + ig] = mult_Gnn_i(&g1[jg1], &g2[jg2], m_Nc);
    }
   }
  }

}

//====================================================================
void Field_G::mult_Field_Gdn(int ex, const Field_G& u1, int ex1,
                                     const Field_G& u2, int ex2)
{
  assert(ex < m_Nex);
  assert(ex1 < u1.nex());
  assert(ex2 < u2.nex());
  assert(u1.nvol() == m_Nvol);
  assert(u2.nvol() == m_Nvol);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));
  double* g1 = const_cast<Field_G*>(&u1)->ptr(m_Nin*(is + m_Nvol*ex1));
  double* g2 = const_cast<Field_G*>(&u2)->ptr(m_Nin*(is + m_Nvol*ex2));

  int Nc2  = 2 * m_Nc;
  int Ndf  = Nc2 * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int ic2 = 0; ic2 < m_Nc; ++ic2){
    for(int ic1 = 0; ic1 < m_Nc; ++ic1){
      int jg2 = ic2 * 2 + ig;
      int jg1 = ic1 * 2 + ig;
      g[2*ic2   + Nc2*ic1 + ig] = mult_Gdn_r(&g1[jg1], &g2[jg2], m_Nc);
      g[2*ic2+1 + Nc2*ic1 + ig] = mult_Gdn_i(&g1[jg1], &g2[jg2], m_Nc);
    }
   }

  }

}

//====================================================================
void Field_G::mult_Field_Gnd(int ex, const Field_G& u1, int ex1,
                                     const Field_G& u2, int ex2)
{
  assert(ex < m_Nex);
  assert(ex1 < u1.nex());
  assert(ex2 < u2.nex());
  assert(u1.nvol() == m_Nvol);
  assert(u2.nvol() == m_Nvol);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));
  double* g1 = const_cast<Field_G*>(&u1)->ptr(m_Nin*(is + m_Nvol*ex1));
  double* g2 = const_cast<Field_G*>(&u2)->ptr(m_Nin*(is + m_Nvol*ex2));

  int Nc2  = 2 * m_Nc;
  int Ndf  = Nc2 * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int ic2 = 0; ic2 < m_Nc; ++ic2){
    for(int ic1 = 0; ic1 < m_Nc; ++ic1){
      int jg2 = ic2 * Nc2 + ig;
      int jg1 = ic1 * Nc2 + ig;
      g[2*ic2   + Nc2*ic1 + ig] = mult_Gnd_r(&g1[jg1], &g2[jg2], m_Nc);
      g[2*ic2+1 + Nc2*ic1 + ig] = mult_Gnd_i(&g1[jg1], &g2[jg2], m_Nc);
    }
   }

  }

}

//====================================================================
void Field_G::mult_Field_Gdd(int ex, const Field_G& u1, int ex1,
                                     const Field_G& u2, int ex2)
{
  assert(ex < m_Nex);
  assert(ex1 < u1.nex());
  assert(ex2 < u2.nex());
  assert(u1.nvol() == m_Nvol);
  assert(u2.nvol() == m_Nvol);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));
  double* g1 = const_cast<Field_G*>(&u1)->ptr(m_Nin*(is + m_Nvol*ex1));
  double* g2 = const_cast<Field_G*>(&u2)->ptr(m_Nin*(is + m_Nvol*ex2));

  int Nc2  = 2 * m_Nc;
  int Ndf  = Nc2 * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int ic2 = 0; ic2 < m_Nc; ++ic2){
    for(int ic1 = 0; ic1 < m_Nc; ++ic1){
      int jg2 = ic2 * Nc2 + ig;
      int jg1 = ic1 * 2   + ig;
      g[2*ic2   + Nc2*ic1 + ig] = mult_Gdd_r(&g1[jg1], &g2[jg2], m_Nc);
      g[2*ic2+1 + Nc2*ic1 + ig] = mult_Gdd_i(&g1[jg1], &g2[jg2], m_Nc);
    }
   }
  }

}

//====================================================================
void Field_G::multadd_Field_Gnn(int ex, const Field_G& u1, int ex1,
                              const Field_G& u2, int ex2, double ff)
{
  assert(ex < m_Nex);
  assert(ex1 < u1.nex());
  assert(ex2 < u2.nex());
  assert(u1.nvol() == m_Nvol);
  assert(u2.nvol() == m_Nvol);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));
  double* g1 = const_cast<Field_G*>(&u1)->ptr(m_Nin*(is + m_Nvol*ex1));
  double* g2 = const_cast<Field_G*>(&u2)->ptr(m_Nin*(is + m_Nvol*ex2));

  int Nc2  = 2 * m_Nc;
  int Ndf  = Nc2 * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int ic2 = 0; ic2 < m_Nc; ++ic2){
    for(int ic1 = 0; ic1 < m_Nc; ++ic1){
      int jg2 = ic2 * 2   + ig;
      int jg1 = ic1 * Nc2 + ig;
      g[2*ic2   + Nc2*ic1 + ig] +=
                         ff* mult_Gnn_r(&g1[jg1], &g2[jg2], m_Nc);
      g[2*ic2+1 + Nc2*ic1 + ig] +=
                         ff* mult_Gnn_i(&g1[jg1], &g2[jg2], m_Nc);
    }
   }
  }

}

//====================================================================
void Field_G::multadd_Field_Gdn(int ex, const Field_G& u1, int ex1,
                              const Field_G& u2, int ex2, double ff)
{
  assert(ex < m_Nex);
  assert(ex1 < u1.nex());
  assert(ex2 < u2.nex());
  assert(u1.nvol() == m_Nvol);
  assert(u2.nvol() == m_Nvol);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));
  double* g1 = const_cast<Field_G*>(&u1)->ptr(m_Nin*(is + m_Nvol*ex1));
  double* g2 = const_cast<Field_G*>(&u2)->ptr(m_Nin*(is + m_Nvol*ex2));

  int Nc2  = 2 * m_Nc;
  int Ndf  = Nc2 * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int ic2 = 0; ic2 < m_Nc; ++ic2){
    for(int ic1 = 0; ic1 < m_Nc; ++ic1){
      int jg2 = ic2 * 2 + ig;
      int jg1 = ic1 * 2 + ig;
      g[2*ic2   + Nc2*ic1 + ig] +=
                         ff* mult_Gdn_r(&g1[jg1], &g2[jg2], m_Nc);
      g[2*ic2+1 + Nc2*ic1 + ig] +=
                         ff* mult_Gdn_i(&g1[jg1], &g2[jg2], m_Nc);
    }
   }
  }

}

//====================================================================
void Field_G::multadd_Field_Gnd(int ex, const Field_G& u1, int ex1,
                                const Field_G& u2, int ex2, double ff)
{
  assert(ex < m_Nex);
  assert(ex1 < u1.nex());
  assert(ex2 < u2.nex());
  assert(u1.nvol() == m_Nvol);
  assert(u2.nvol() == m_Nvol);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));
  double* g1 = const_cast<Field_G*>(&u1)->ptr(m_Nin*(is + m_Nvol*ex1));
  double* g2 = const_cast<Field_G*>(&u2)->ptr(m_Nin*(is + m_Nvol*ex2));

  int Nc2  = 2 * m_Nc;
  int Ndf  = Nc2 * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int ic2 = 0; ic2 < m_Nc; ++ic2){
    for(int ic1 = 0; ic1 < m_Nc; ++ic1){
      int jg2 = ic2 * Nc2 + ig;
      int jg1 = ic1 * Nc2 + ig;
      g[2*ic2   + Nc2*ic1 + ig] +=
                         ff* mult_Gnd_r(&g1[jg1], &g2[jg2], m_Nc);
      g[2*ic2+1 + Nc2*ic1 + ig]
                      += ff* mult_Gnd_i(&g1[jg1], &g2[jg2], m_Nc);
    }
   }
  }

}

//====================================================================
void Field_G::multadd_Field_Gdd(int ex, const Field_G& u1, int ex1,
                                const Field_G& u2, int ex2, double ff)
{
  assert(ex < m_Nex);
  assert(ex1 < u1.nex());
  assert(ex2 < u2.nex());
  assert(u1.nvol() == m_Nvol);
  assert(u2.nvol() == m_Nvol);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));
  double* g1 = const_cast<Field_G*>(&u1)->ptr(m_Nin*(is + m_Nvol*ex1));
  double* g2 = const_cast<Field_G*>(&u2)->ptr(m_Nin*(is + m_Nvol*ex2));

  int Nc2  = 2 * m_Nc;
  int Ndf  = Nc2 * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int ic2 = 0; ic2 < m_Nc; ++ic2){
    for(int ic1 = 0; ic1 < m_Nc; ++ic1){
      int jg2 = ic2 * Nc2 + ig;
      int jg1 = ic1 * 2   + ig;
      g[2*ic2   + Nc2*ic1 + ig] +=
                        ff* mult_Gdd_r(&g1[jg1], &g2[jg2], m_Nc);
      g[2*ic2+1 + Nc2*ic1 + ig] +=
                        ff* mult_Gdd_i(&g1[jg1], &g2[jg2], m_Nc);
    }
   }
  }

}

//====================================================================
void Field_G::at_Field_G(int ex)
{
  assert(ex < m_Nex);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));

  int Ndf  = 2 * m_Nc * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int a = 0; a < m_Nc; ++a){
      for(int b = a+1; b < m_Nc; ++b){

        double re =  g[2*(m_Nc*a+b)  +ig] - g[2*(m_Nc*b+a)  +ig];
        double im =  g[2*(m_Nc*a+b)+1+ig] + g[2*(m_Nc*b+a)+1+ig];

        g[2*(m_Nc*a+b)  +ig] =  0.5 * re;
        g[2*(m_Nc*a+b)+1+ig] =  0.5 * im;

        g[2*(m_Nc*b+a)  +ig] = -0.5 * re;
        g[2*(m_Nc*b+a)+1+ig] =  0.5 * im;
      }
    }
    double tr=0.0;
    for(int cc = 0; cc < m_Nc; ++cc){
      tr += g[2*(m_Nc*cc+cc)+1+ig];
    }
    tr = tr/m_Nc;
    for(int cc = 0; cc < m_Nc; ++cc){
      g[2*(m_Nc*cc+cc)  +ig]  = 0.0;
      g[2*(m_Nc*cc+cc)+1+ig] -= tr;
    }

  }

}

//====================================================================
void Field_G::ah_Field_G(int ex)
{
  assert(ex < m_Nex);

  int nth = ThreadManager_OpenMP::get_num_threads();
  int ith = ThreadManager_OpenMP::get_thread_id();

  int is  = m_Nvol * ith / nth;
  int ns  = m_Nvol * (ith+1) / nth - is;

  double* g  = ptr(m_Nin*(is + m_Nvol*ex));

  int Ndf  = 2 * m_Nc * m_Nc;

  for(int site = 0; site < ns; ++site){
    int ig  = Ndf  * site;

    for(int a = 0; a < m_Nc; ++a){
      for(int b = a; b < m_Nc; ++b){

        double re =  g[2*(m_Nc*a+b)  +ig] - g[2*(m_Nc*b+a)  +ig];
        double im =  g[2*(m_Nc*a+b)+1+ig] + g[2*(m_Nc*b+a)+1+ig];

        g[2*(m_Nc*a+b)  +ig] =  0.5 * re;
        g[2*(m_Nc*a+b)+1+ig] =  0.5 * im;

        g[2*(m_Nc*b+a)  +ig] = -0.5 * re;
        g[2*(m_Nc*b+a)+1+ig] =  0.5 * im;
      }
    }

  }

}

//====================================================================
//============================================================END=====
