/*!
        @file    $Id:: shiftField_lex.h #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru) 
                 $LastChangedBy: matufuru $

        @date    $LastChangedDate:: 2013-10-20 19:28:34 #$

        @version $LastChangedRevision: 972 $
*/


#ifndef SHIFTFIELD_LEX_INCLUDED
#define SHIFTFIELD_LEX_INCLUDED

#include <iostream>

#include "communicator.h"
#include "commonParameters.h"

#include "field.h"
#include "index_lex.h"

//! Methods to shift the field in lexical site index.
/*!
   This class defines the methods which shift a given Field
   instance in specified direction.
   The foreward shift means, e.g. in x-direction,
   v(site) = w(site-\hat{mu}), where v is the shifted field
   (output, first argument) and w the original field (input,
   second argument).
   Names of private functions might be still confusing.
                                   [25 Dec 2011 H.Matsufuru]
 */
class ShiftField_lex{

 private:
  int Nx, Ny, Nz, Nt;
  int m_Nin;          //!< inner (on-site) degree of freedom.
  Index_lex idx;

  //! pimple prescription for ShiftField_lex class implementation.
  class ShiftField_lex_impl;
  ShiftField_lex_impl *m_impl;

public:
  ShiftField_lex();
  ShiftField_lex(const int Nin);

  ~ShiftField_lex();

  void  forward(Field&, const Field&, const int mu);
  void backward(Field&, const Field&, const int mu);

  void  forward(Field&, const Field&, const int bc, const int mu);
  void backward(Field&, const Field&, const int bc, const int mu);

  void  forward_ex(Field&, int, const Field&, int, const int mu);
  void backward_ex(Field&, int, const Field&, int, const int mu);

  void  forward_ex(Field&, int, const Field&, int, const int bc, const int mu);
  void backward_ex(Field&, int, const Field&, int, const int bc, const int mu);

 private:
  void up_x(Field*, const Field*, const int bc);
  void up_y(Field*, const Field*, const int bc);
  void up_z(Field*, const Field*, const int bc);
  void up_t(Field*, const Field*, const int bc);
  void dn_x(Field*, const Field*, const int bc);
  void dn_y(Field*, const Field*, const int bc);
  void dn_z(Field*, const Field*, const int bc);
  void dn_t(Field*, const Field*, const int bc);

  void up_x_ex(Field*, int, const Field*, int, const int bc);
  void up_y_ex(Field*, int, const Field*, int, const int bc);
  void up_z_ex(Field*, int, const Field*, int, const int bc);
  void up_t_ex(Field*, int, const Field*, int, const int bc);
  void dn_x_ex(Field*, int, const Field*, int, const int bc);
  void dn_y_ex(Field*, int, const Field*, int, const int bc);
  void dn_z_ex(Field*, int, const Field*, int, const int bc);
  void dn_t_ex(Field*, int, const Field*, int, const int bc);

};

#endif
