/*!
         @file    $Id:: main.cpp #$
 
         @brief
 
         @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                  $LastChangedBy: matufuru $
 
         @date    $LastChangedDate:: 2014-06-03 00:18:24 #$
 
         @version $LastChangedRevision: 1056 $
*/

#include <iostream>
#include <fstream>
#include <string>
using std::string;
#include <valarray>
using std::valarray;

#include "configure.h"
#include "defs.h"
#include "commonParameters.h"

#include "bridgeIO.h"
using Bridge::vout;

#include "threadManager_OpenMP.h"

#include "timer.h"

//#include "main.h"
namespace Test_Spectrum_Domainwall {
  int hadron_2ptFunction(void);
}

//====================================================================
int main(int argc, char *argv[])
{
  // output setup.
  Bridge::vout.init(std::cout);
  Bridge::vout.ildg_init(std::cout);
  Bridge::VerboseLevel vl = Bridge::GENERAL;

  // communicator setup.
  Communicator::init(&argc, &argv);

  // lattice, grid, and thread parameters

  valarray<int> lattice_size(4);
  /*
  lattice_size[0] = 32;
  lattice_size[1] = 32;
  lattice_size[2] = 32;
  lattice_size[3] = 32;
  */
  lattice_size[0] = 16;
  lattice_size[1] = 16;
  lattice_size[2] = 16;
  lattice_size[3] = 32;
  /*
  lattice_size[0] = 8;
  lattice_size[1] = 8;
  lattice_size[2] = 8;
  lattice_size[3] = 16;
  */
  /*
  lattice_size[0] = 4;
  lattice_size[1] = 4;
  lattice_size[2] = 4;
  lattice_size[3] = 8;
  */
  valarray<int> grid_size(4);
// 256 node
/*
  grid_size[0] = 2;
  grid_size[1] = 4;
  grid_size[2] = 4;
  grid_size[3] = 8;
*/
// 128 node
  grid_size[0] = 2;
  grid_size[1] = 2;
  grid_size[2] = 4;
  grid_size[3] = 8;
// 64 node
/*
  grid_size[0] = 2;
  grid_size[1] = 2;
  grid_size[2] = 4;
  grid_size[3] = 4;
*/
// 32 node
/*
  grid_size[0] = 2;
  grid_size[1] = 2;
  grid_size[2] = 2;
  grid_size[3] = 4;
*/
// 4 node
  /*
  grid_size[0] = 1;
  grid_size[1] = 1;
  grid_size[2] = 2;
  grid_size[3] = 2;
  */
// 1 node
  /*
  grid_size[0] = 1;
  grid_size[1] = 1;
  grid_size[2] = 1;
  grid_size[3] = 1;
  */

  //  int number_of_thread = 1;
  int number_of_thread = 2;
  //  int number_of_thread = 4;
  //  int number_of_thread = 16;
  //  int number_of_thread = 32;

  // common parameters setup.
  CommonParameters::init(lattice_size, grid_size);
  Communicator::setup();

  ThreadManager_OpenMP::init(number_of_thread);

  int Ndim = CommonParameters::Ndim();
  int Nvol = CommonParameters::Nvol();


  //- print input parameters
  vout.general(vl, "Main: input parameters\n");
  for (int mu = 0; mu < Ndim; ++mu) {
    vout.general(vl, "  lattice_size[%d] = %d\n", mu, lattice_size[mu]);
    vout.general(vl, "  grid_size[%d]    = %d\n", mu, grid_size[mu]);
  }
  vout.general(vl, "  number of thread = %d\n", number_of_thread);
  vout.general(vl, "\n");

  //- timestamp (starting time)
  Timer *timer = new Timer;
  timer->timestamp();
  timer->start();

  // main part.
  Test_Spectrum_Domainwall::hadron_2ptFunction();

  //- timestamp (end time)
  timer->report();
  timer->timestamp();

  // ####  tydy up  ####
  delete timer;

  Communicator::finalize();

  return EXIT_SUCCESS;
}

//====================================================================
//============================================================END=====
