/*!
        @file    $Id:: fprop_Domainwall_4d.cpp #$

        @brief

        @author  <Hideo Matsufuru> hideo.matsufuru@kek.jp(matsufuru)
                 $LastChangedBy: namekawa $

        @date    $LastChangedDate:: 2013-05-29 16:29:31 #$

        @version $LastChangedRevision: 889 $
*/

#include "fprop_Domainwall_4d.h"

const std::string Fprop_Domainwall_4d::class_name = "Fprop_Domainwall_4d";

//====================================================================
void Fprop_Domainwall_4d::set_config(Field *U)
{
  m_fopr_dw->set_config(U);
}

//====================================================================
void Fprop_Domainwall_4d::invert_D(Field& xq, const Field& b,
                                           int& Nconv, double& diff)
{
  // ####  parameter setup  ####
  int Nin  = m_fopr_dw->field_nin();
  int Nvol = m_fopr_dw->field_nvol();
  int Nex  = m_fopr_dw->field_nex();

  assert(b.nin() == Nin);
  assert(b.nvol() == Nvol);
  assert(b.nex() == 1);

  // ####  object setup  ####
  Field bt(Nin, Nvol, 1);
  Field xt(Nin, Nvol, 1), xt2(Nin, Nvol, 1);
  Field b_5d(Nin, Nvol, Nex), bt_5d(Nin, Nvol, Nex);
  Field xq_5d(Nin, Nvol, Nex);

  xq.clear();  //  xq = 0.0;

  for (int ch = -1; ch <= 1; ch = ch + 2) {
    //    bt = m_fopr_dw->mult_chproj_4d(b, ch);
    m_fopr_dw->mult_chproj_4d(bt, b, ch);
    if (bt.norm2() < 1.e-14) continue;

    bt_5d = 0.0;
    if (ch == -1) {
      bt_5d.setpart_ex(Nex - 1, bt, 0);
    } else {
      bt_5d.setpart_ex(0, bt, 0);
    }

    m_fopr_dw->set_mode("Ddag");
    //  b_5d = m_fopr_dw->mult(bt_5d);
    m_fopr_dw->mult(b_5d, bt_5d);

    vout.paranoiac(m_vl, " b_5d norm2 = %14.6e\n", b_5d.norm2());

    m_fopr_dw->set_mode("DdagD");
    m_solver->solve(xq_5d, b_5d, Nconv, diff);
    vout.detailed(m_vl, "  chirality: %2d\n", ch);
    vout.detailed(m_vl, "    Nconv = %d,  diff = %16.8e\n", Nconv, diff);

    xt.setpart_ex(0, xq_5d, 0);
    //  xt2 = m_fopr_dw->mult_chproj_4d(xt, -1);
    m_fopr_dw->mult_chproj_4d(xt2, xt, -1);
    xq.addpart_ex(0, xt2, 0);

    xt.setpart_ex(0, xq_5d, Nex - 1);
    // xt2 = m_fopr_dw->mult_chproj_4d(xt, +1);
    m_fopr_dw->mult_chproj_4d(xt2, xt, +1);
    xq.addpart_ex(0, xt2, 0);
  }
}


//====================================================================
void Fprop_Domainwall_4d::invert_DdagD(Field& xq, const Field& b, int& Nconv, double& diff)
{
  //- Not implemented, yet.
}


//====================================================================
//============================================================END=====
